/* @(#)xheader.c	1.15 02/06/10 Copyright 2001 J. Schilling */
#ifndef lint
static	char sccsid[] =
	"@(#)xheader.c	1.15 02/06/10 Copyright 2001 J. Schilling";
#endif
/*
 *	Handling routines to read/write, parse/create
 *	POSIX.1-2001 extended archive headers
 *
 *	Copyright (c) 2001 J. Schilling
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <mconfig.h>
#include <stdio.h>
#include <stdxlib.h>
#include <unixstd.h>	/* getpagesize() */
#include <libport.h>	/* getpagesize() */
#include "star.h"
#include "props.h"
#include "table.h"
#include <dirdefs.h>
#include <standard.h>
#include <strdefs.h>
#define	__XDEV__	/* Needed to activate _dev_major()/_dev_minor() */
#include <device.h>
#include <schily.h>
#include "starsubs.h"
#include "movearch.h"

#define	MAX_UNAME	64	/* The maximum length of a user/group name */

/*
 * Flags for gen_text()
 */
#define	T_ADDSLASH	1	/* Add slash to the argument	*/
#define	T_UTF8		2	/* Convert arg to UTF-8 coding	*/

typedef	void (*function)	__PR((FINFO *, char *, char *, int));

typedef struct {
	char		*x_name;
	int		x_namelen;
	function	x_func;
	int		x_flag;
} xtab_t;

EXPORT	void	xbinit		__PR((void));
LOCAL	void	xbgrow		__PR((int newsize));
EXPORT	void	info_to_xhdr	__PR((FINFO * info, TCB * ptb));
LOCAL	void	gen_xtime	__PR((char *keyword, time_t sec, Ulong nsec));
LOCAL	void	gen_number	__PR((char *keyword, Llong arg));
LOCAL	void	gen_text	__PR((char *keyword, char *arg, int alen, Uint flags));
LOCAL	int	len_len		__PR((int len));
LOCAL	xtab_t	*lookup		__PR((char *cmd, int clen, xtab_t *cp));
EXPORT	int	tcb_to_xhdr	__PR((TCB * ptb, FINFO * info));
LOCAL	BOOL	get_xtime	__PR((char *keyword, char *arg, time_t *secp,
								Ulong *nsecp));
LOCAL	void	get_atime	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_ctime	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_mtime	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	BOOL	get_number	__PR((char *keyword, char *arg, Llong *llp));
LOCAL	void	get_uid		__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_gid		__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_uname	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_gname	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_path	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_lpath	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_size	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_major	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_minor	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_dev		__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_ino		__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_nlink	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_filetype	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_acl_access	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_acl_default	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_xfflags	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_dir		__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	get_dummy	__PR((FINFO *info, char *keyword, char *arg, int len));
LOCAL	void	bad_utf8	__PR((char *keyword, char *arg));

/*
 * Convert unsigned integer into a decimal string.
 * The parameter 'val' is of the needed size (basic type), so it is as fast
 * as possible.
 *
 * The string is written starting from the end backwards for best speed.
 */
LOCAL	Uchar	dtab[] = "0123456789";

#define	ui2decimal(val, np)	do {					      \
					*--(np) = dtab[(val) % (unsigned)10]; \
					(val) = (val) / (unsigned)10;	      \
				} while ((val) > 0)

#define	scopy(to, from)		while ((*(to)++ = *(from)++) != '\0');


LOCAL	char	*xbuf;	/* Space used to prepare I/O from/to extended headers */
LOCAL	int	xblen;	/* the length of the buffer for the extended headers */
LOCAL	int	xbidx;	/* The index where we start to prepare next entry    */

/*
 * Important for the correctness of ben_number(): the maximum length entry
 * currently is "SCHILY.acl.default" and it's length is 19 chars. As long as
 * we stay <= 55 chars for the keyword, a number entry will fit into 100 chars.
 */
LOCAL xtab_t xtab[] = {
			{ "atime",		5, get_atime,	0	},
			{ "ctime",		5, get_ctime,	0	},
			{ "mtime",		5, get_mtime,	0	},

			{ "uid",		3, get_uid,	0	},
			{ "uname",		5, get_uname,	0	},
			{ "gid",		3, get_gid,	0	},
			{ "gname",		5, get_gname,	0	},

			{ "path",		4, get_path,	0	},
			{ "linkpath",		8, get_lpath,	0	},

			{ "size",		4, get_size,	0	},

			{ "charset",		7, get_dummy,	0	},
			{ "comment",		7, get_dummy,	0	},

			{ "SCHILY.devmajor",   15, get_major,	0	},
			{ "SCHILY.devminor",   15, get_minor,	0	},

#ifdef	USE_ACL
			{ "SCHILY.acl.access", 17, get_acl_access,0	},
			{ "SCHILY.acl.default",18, get_acl_default,0	},
#else
/*
 * We don't want star to complain about unknown extended headers when it
 * has been compiled without ACL support.
 */
			{ "SCHILY.acl.access", 17, get_dummy,	0	},
			{ "SCHILY.acl.default",18, get_dummy,	0	},
#endif

#ifdef	USE_FFLAGS
			{ "SCHILY.fflags",     13, get_xfflags,	0	},
#else
/*
 * We don't want star to complain about unknown extended headers when it
 * has been compiled without extended file flag support.
 */
			{ "SCHILY.fflags",     13, get_dummy,	0	},
#endif
			{ "SCHILY.dev",	       10, get_dev,	0	},
			{ "SCHILY.ino",	       10, get_ino,	0	},
			{ "SCHILY.nlink",      12, get_nlink,	0	},
			{ "SCHILY.filetype",   15, get_filetype,0	},
			{ "SCHILY.tarfiletype",18, get_filetype,0	},

			{ "SCHILY.dir",	       10, get_dir,	0	},
			{ "SCHILY.ddev",       11, get_dummy,	0	},
			{ "SCHILY.dino",       11, get_dummy,	0	},

			{ "SUN.devmajor",      12, get_major,	0	},
			{ "SUN.devminor",      12, get_minor,	0	},

			{ NULL,			0, NULL,	0	}};

/*
 * Initialize the growable buffer used for reading the extended headers
 */
EXPORT void
xbinit()
{
	xbuf = __malloc(1, "growable xheader");
	xblen = 1;
}

/*
 * Grow the growable buffer used for reading the extended headers
 */
LOCAL void
xbgrow(newsize)
	int	newsize;
{
	char	*newbuf;
	int	i;
	int	ps = getpagesize();

	/*
	 * grow in pagesize units
	 */
	for (i = 0; i < newsize; i += ps)
		;
	newsize = i + xblen;
	newbuf = __realloc(xbuf, newsize, "growable xheader");
	xbuf = newbuf;
	xblen = newsize;
}

/*
 * Prepare and write out the extended header
 */
EXPORT void
info_to_xhdr(info, ptb)
	register FINFO	*info;
	register TCB	*ptb;
{
	FINFO	finfo;
	TCB	*xptb;
	move_t	move;
	char	name[MAX_UNAME+1];
extern	long	hdrtype;
extern	BOOL	dodump;

	fillbytes((char *)&finfo, sizeof(finfo), '\0');

	/*
	 * Unless we really don't want extended sub-second resolution
	 * timestamps or a specific selection of timestams has been set up,
	 * include all times (atime/ctime/mtime) if we need to include extended
	 * headers at all.
	 */
	if ((info->f_xflags & (XF_ATIME|XF_CTIME|XF_MTIME|XF_NOTIME)) == 0)
		info->f_xflags |= (XF_ATIME|XF_CTIME|XF_MTIME);

/*#define	DEBUG_XHDR*/
#ifdef	DEBUG_XHDR
	info->f_xflags = 0xffffffff;
#endif

	xbidx = 0;	/* Reset xbuffer index to start of buffer. */

	if (info->f_xflags & XF_ATIME)
		gen_xtime("atime", info->f_atime, info->f_ansec);
	if (info->f_xflags & XF_CTIME)
		gen_xtime("ctime", info->f_ctime, info->f_cnsec);
	if (info->f_xflags & XF_MTIME)
		gen_xtime("mtime", info->f_mtime, info->f_mnsec);

	if (info->f_xflags & XF_UID)
		gen_number("uid", (Llong)info->f_uid);
	if (info->f_xflags & XF_GID)
		gen_number("gid", (Llong)info->f_gid);

	if (info->f_xflags & XF_UNAME) {
		nameuid(name, sizeof(name)-1, info->f_uid);
		gen_text("uname", name, -1, T_UTF8);
	}
	if (info->f_xflags & XF_GNAME) {
		namegid(name, sizeof(name)-1, info->f_gid);
		gen_text("gname", name, -1, T_UTF8);
	}

	if (info->f_xflags & XF_PATH) {
		gen_text("path", info->f_name, -1,
			(info->f_flags & F_ADDSLASH) != 0 ?
				(T_ADDSLASH|T_UTF8) : T_UTF8);
	}

	if (info->f_xflags & XF_LINKPATH && info->f_lnamelen)
		gen_text("linkpath", info->f_lname, -1, T_UTF8);

	if (info->f_xflags & XF_SIZE)
		gen_number("size", (Llong)info->f_size);

	if (H_TYPE(hdrtype) == H_SUNTAR) {
		if (info->f_xflags & XF_DEVMAJOR)
			gen_number("SUN.devmajor", (Llong)info->f_rdevmaj);
		if (info->f_xflags & XF_DEVMINOR)
			gen_number("SUN.devminor", (Llong)info->f_rdevmin);
	} else {
		if (info->f_xflags & XF_DEVMAJOR)
			gen_number("SCHILY.devmajor", (Llong)info->f_rdevmaj);
		if (info->f_xflags & XF_DEVMINOR)
			gen_number("SCHILY.devminor", (Llong)info->f_rdevmin);
	}

#ifdef	USE_ACL
	/*
	 * POSIX Access Control Lists, currently supported e.g. by Linux.
	 * Solaris ACLs have been converted to POSIX before.
	 */
	if (info->f_xflags & XF_ACL_ACCESS) {
		gen_text("SCHILY.acl.access", info->f_acl_access, -1, T_UTF8);
	}

	if (info->f_xflags & XF_ACL_DEFAULT) {
		gen_text("SCHILY.acl.default", info->f_acl_default, -1, T_UTF8);
	}
#endif  /* USE_ACL */

#ifdef	USE_FFLAGS
	if (info->f_xflags & XF_FFLAGS) {
extern char	*textfromflags	__PR((FINFO *, char *));

		char	fbuf[512];
		gen_text("SCHILY.fflags", textfromflags(info, fbuf), -1, 0);
	}
#endif

	if (dodump) {
		gen_number("SCHILY.dev", (Llong)info->f_dev);
		gen_number("SCHILY.ino", (Llong)info->f_ino);
		gen_number("SCHILY.nlink", (Llong)info->f_nlink);
		gen_text("SCHILY.filetype", XTTONAME(info->f_rxftype), -1, 0);
#ifdef	__needed__
		if (info->f_rxftype != info->f_xftype)
			gen_text("SCHILY.tarfiletype", XTTONAME(info->f_xftype), -1, 0);
#endif
		if (is_dir(info) && info->f_dir) {
			gen_text("SCHILY.dir", info->f_dir, info->f_dirlen, T_UTF8);
		}
	}

	xptb = (TCB *)get_block();
	finfo.f_flags |= F_TCB_BUF;
	filltcb(xptb);
/*	strcpy(xptb->dbuf.t_name, ptb->dbuf.t_name);*/
	strcpy(xptb->dbuf.t_name, "././@PaxHeader");
	finfo.f_rsize = finfo.f_size = xbidx;
	finfo.f_xftype = props.pr_xc;
	info_to_tcb(&finfo, xptb);
	xptb->dbuf.t_linkflag = props.pr_xc;
	write_tcb(xptb, &finfo);

	move.m_data = xbuf;
	move.m_size = finfo.f_size;
	move.m_flags = 0;
	cr_file(&finfo, vp_move_to_arch, &move, 0, "moving extended header");
}

/*
 * Create a time string with sub-second granularity.
 *
 * <length> <time-name-spec>=<seconds>.<nano-seconds>\n
 *
 *	00 atime=<seconds>.123456789\n
 *	00 ctime=<seconds>.123456789\n
 *	00 mtime=<seconds>.123456789\n
 *
 *	00 mtime=.123456789\n
 *	.123456789.123456789
 *
 * As we always emmit 9 digits for the sub-second part, the
 * length of this entry is always more than 20 and less than 100.
 * The size of an entry is 20 + the size of the "seconds" part of the entry.
 * With 64 bit unsigned numbers, the "seconds" part will be 20 char at max.
 * We may safely fill in the two digit <length> later, when we know the value.
 *
 * The code is highly optimised because in PAX (extended TAR) mode we are
 * spending a significant amount of user CPU time in it.
 */
LOCAL void
gen_xtime(keyword, sec, nsec)
		 char	*keyword;
	register time_t	sec;
	register Ulong	nsec;
{
		 char	nb[32];
	register char	*p;
	register char	*np;
	register int	len;

	if ((xbidx + 100) > xblen)
		xbgrow(100);

	/*
	 * The following code is equivalent to:
	 *
	 * sprintf(p, "%d %s=%lld.%9.9ld\n", length, keyword, (Llong)sec, nsec)
	 *
	 * But is twice as fast.
	 */
	len = 20;		/* Base length, second field must be added   */
	p = &xbuf[xbidx+2];	/* point past length field		     */
	/*
	 * Fill in " ?time="
	 */
	*p++ = ' ';
	*p++ = keyword[0]; *p++ = 't'; *p++ = 'i'; *p++ = 'm'; *p++ = 'e';
	*p++ = '=';
	if (sec < 0) {		/* Time is before 01.01.1970 ...	     */
		*p++ = '-';
		sec = -sec;
		len += 1;
	}

	np = &nb[sizeof(nb)-1];	/* Point to end of number string buffer	     */
	*np = '\0';		/* Null terminate			     */
	ui2decimal(sec, np);	/* Convert to unsigned decimal string	     */
	len += &nb[sizeof(nb)-1] - np;

	scopy(p, np);		/* Copy number string buffer to xheader	     */
	--p,			/* Overshoot compensation		     */
	*p++ = '.';		/* Decimal point between secs and nsecs	     */

	np = &p[10];		/* Point to end of <nsec> string part	     */
	*np = '\0';		/* Null terminate			     */
	*--np = '\n';		/* Newline at end of line		     */
	ui2decimal(nsec, np);	/* Convert to unsigned decimal string	     */
	while (np > p)
		*--np = '0';	/* Left fill with '0' to 9 digits	     */

	np = &xbuf[xbidx+2];	/* Point to end of length string part	     */
	xbidx += len;		/* 'len' is trashed in ui2decimal()	     */
	ui2decimal(len, np);	/* Convert to unsigned decimal string	     */
}

/*
 * Create a generic number string.
 *
 * <length> <name-spec>=<value>\n
 *
 * The length of this entry is always less than 100 chars if the length of the
 * 'name-spec' is less than 75 chars (the maximum length of a 64 bit number in
 * decimal is 20 digits). Even in case of a 128 bit number length will be less
 * than 100 chars if the length of 'name-spec' is less than 55 chars.
 *
 * The code is highly optimised because in dump mode when using extended TAR
 * headers with additional fields, we are spending a significant amount of
 * user CPU time in it.
 */
LOCAL void
gen_number(keyword, arg)
	register char	*keyword;
	register Llong	arg;
{
		 char	nb[64];	/* 41 is enough for unsigned 128 bit ints    */
	register char	*p;
	register char	*np;
	register int	len;
	register int	i;

	if ((xbidx + 100) > xblen)
		xbgrow(100);

	/*
	 * The following code is equivalent to:
	 *
	 * sprintf(p, "%d %s=%lld\n", length, keyword, arg);
	 *
	 * But is twice as fast.
	 */
	np = &nb[sizeof(nb)-1];	/* Point to end of number string buffer	     */
	*np = '\0';		/* Null terminate			     */
	ui2decimal(arg, np);	/* Convert to unsigned decimal string	     */

	len = strlen(keyword);	/* Compute the length, start with strlen(kw) */

	len += &nb[sizeof(nb)-1] - np;	/* Add strlen(number)		     */
	len += 2 + 3;		/* Add 2 for strlen(len) and 3 for " =\n"    */

	if (len < 10) {		/* If < 10, the len field is 1 digit only    */
		len -= 1;	/* This happens when strlen(keyword) is < 4  */
		i = 1;		/* and number is one digit only.	     */
	} else {
		i = 2;
	}
	p = &xbuf[xbidx+i];	/* Point past length field		     */
	*p++ = ' ';		/* Fill in space after length field	     */
	scopy(p, keyword);	/* Copy keyword into to xheader		     */
	--p,			/* Overshoot compensation		     */
	*p++ = '=';		/* Fill in '=' after keyword field	     */

	scopy(p, np);		/* Copy number string buffer to xheader	     */
	*--p = '\n';		/* Newline at end of line		     */

	np = &xbuf[xbidx+i];	/* Point to end of length string part	     */
	xbidx += len;		/* 'len' is trashed in ui2decimal()	     */
	ui2decimal(len, np);	/* Convert to unsigned decimal string	     */
}

/*
 * Create a generic text string in UTF-8 coding.
 *
 * <length> <name-spec>=<value>\n
 *
 * This function will have to carefully check for the resultant length
 * and thus compute the total length in advance. If the rare case that the
 * UTF-8 conversion changes the length so much that the length of the length
 * string will be different from the estimated value, we need to move whole
 * text by one.
 *
 * The code is highly optimised because in dump mode when using extended TAR
 * headers with additional fields, we may need to copy directory listings
 * that are longer then 100000 bytes.
 */
LOCAL void
gen_text(keyword, arg, alen, flags)
	register char	*keyword;
		 char	*arg;
		 int	alen;
		 Uint	flags;
{
	register char	*p;
	register char	*np;
	register int	len;
	register int	i;
	register int	llen;
	register int	olen;

	/*
	 * The following code is equivalent to:
	 *
	 * sprintf(p, "%d %s=%s\n", length, keyword, arg);
	 *
	 * But avoids copying if possible.
	 */
	if ((len = alen) == -1)
		len = strlen(arg);
	alen = len;
	if (flags & T_ADDSLASH)		/* only used if 'path' is a dir	     */
		len++;

	len += strlen(keyword) + 3;	/* + ' ' + '=' + '\n'		     */
	olen = len;
	len += llen = len_len(len);	/* add length of length string	     */

	i = len;
	if (flags & T_UTF8)
		i *= 2;			/* UTF-8 Factor may be up to 6!	     */
	if ((xbidx + i) > xblen)
		xbgrow(i);


	p = &xbuf[xbidx+llen];	/* Point past length field		     */
	*p++ = ' ';		/* Fill in space after length field	     */
	scopy(p, keyword);	/* Copy keyword into to xheader		     */
	--p,			/* Overshoot compensation		     */
	*p++ = '=';		/* Fill in '=' after keyword field	     */

	if (flags & T_UTF8) {
		i = to_utf8l((Uchar *)p, (Uchar *)arg, alen);
		p += i + 1;
	} else {
		p = movebytes(arg, p, alen);
		p++;
	}

	if (flags & T_ADDSLASH) {/* only used if 'path' is a dir	     */
		i++;		/* String length increases by '/'	     */
		*--p = '/';	/* Slash at end of string		     */
		*++p = '\n';	/* Newline at end of line		     */
	} else {
		*--p = '\n';	/* Newline at end of line		     */
	}
	i = p - &xbuf[xbidx] + 1;/* Total string length			     */
	i -= llen;		/* Subtract old length length value	     */

	if (olen != i) {	/* Length without length string changed	     */
		olen = llen;	/* Save old length string length	     */
		llen = len_len(i);	/* New length of length string	     */
		if (olen != llen) {	/* We need to move the whole text    */
			olen = llen - olen;
			p = &xbuf[xbidx+llen];
			movebytes(p, p+olen, len);
		}
		len = i + llen;
	}

	np = &xbuf[xbidx+llen];	/* Point to end of length string part	     */
	*np = ' ';		/* May have been overwritten by movebytes()  */
	xbidx += len;		/* 'len' is trashed in ui2decimal()	     */
	ui2decimal(len, np);	/* Convert to unsigned decimal string	     */
}

LOCAL int
len_len(len)
	register int	len;
{
	if (len <= 8)
		return (1);
	if (len <= 97)
		return (2);
	if (len <= 996)
		return (3);
	if (len <= 9995)
		return (4);
	if (len <= 99994)
		return (5);
	if (len <= 999993)
		return (6);
	if (len <= 9999992)
		return (7);
	if (len <= 99999991)
		return (8);
	if (len <= 999999990)
		return (9);

	return (10);
}

/*
 * Lookup command in command tab
 */
LOCAL xtab_t *
lookup(cmd, clen, cp)
	register char	*cmd;
	register int	clen;
	register xtab_t	*cp;
{
	for (; cp->x_name; cp++) {
		if (clen != cp->x_namelen)
			continue;

		if ((*cmd == *cp->x_name) &&
		    strcmp(cmd, cp->x_name) == 0) {
			return (cp);
		}
	}
	return ((xtab_t *)NULL);
}

/*
 * Read extended POSIX.1-2001 header and parse the content.
 */
EXPORT int
tcb_to_xhdr(ptb, info)
	register TCB	*ptb;
	register FINFO	*info;
{
	register xtab_t	*cp;
	register char	*keyword;
	register char 	*arg;
	register char	*p;
	register char 	*ep;
		 move_t	move;
		 Ullong	ull;
		 long	length;

#ifdef	XH_DEBUG
error("Block: %lld\n", tblocks());
#endif
	/*
	 * File size is strlen of extended header
	 */
	stolli(ptb->dbuf.t_size, &ull);
	info->f_size = ull;
	info->f_rsize = (off_t)info->f_size;
	/*
	 * Reset xbidx to make xbgrow() work correctly for our case.
	 */
	xbidx = 0;
	if ((info->f_size+1) > xblen)
		xbgrow(info->f_size+1);

	move.m_data = xbuf;
	move.m_flags = 0;
	if (xt_file(info, vp_move_from_arch, &move, 0,
						"moving extended header") < 0) {
		die(EX_BAD);
	}

#ifdef	XH_DEBUG
error("Block: %lld\n", tblocks());
error("xbuf: '%s'\n", xbuf);
#endif

	p = xbuf;
	ep = p + ull;
	while (p < ep) {
		register int klen;

		keyword = astolb(p, &length, 10);
		if (*keyword != ' ') {
			errmsgno(EX_BAD, "Syntax error in extended header.\n");
			break;
		}
		keyword++;
		arg = strchr(keyword, '=');
		klen = arg - keyword;
		*arg++ = '\0';
		*(p + length -1) = '\0';	/* Kill new-line character */

		if ((cp = lookup(keyword, klen, xtab)) != NULL) {
			(*cp->x_func)(info, keyword, arg, length - (arg-p));
		} else {
			errmsgno(EX_BAD,
			    "Unknown extended header keyword '%s' ignored.\n",
				keyword);
		}
		p += length;
	}
	return (get_tcb(ptb));
}

/*
 * generic function to read args that hold times
 *
 * The time may either be in second resolution or in sub-second resolution.
 * In the latter case the second fraction and the sub second fraction
 * is separated by a dot ('.').
 */
LOCAL BOOL
get_xtime(keyword, arg, secp, nsecp)
	char	*keyword;
	char	*arg;
	time_t	*secp;
	Ulong	*nsecp;
{
	Llong	ll;
	long	l;
	char	*p;

	p = astollb(arg, &ll, 10);
	if (*p == '\0') {		/* time has second resolution only */
		*secp = ll;
		*nsecp = 0;
		return (TRUE);
	} else if (*p == '.') {		/* time has sub-second resolution */
		*secp = ll;
		if (strlen(++p) > 9)	/* if resolution is better then nano- */ 
			p[9] = '\0';	/* seconds kill rest of line as we    */
		p = astolb(p, &l, 10);	/* don't understand more.	      */
		if (*p == '\0') {	/* number read correctly	      */
			if (l >= 0) {
				*nsecp = l;
				return (TRUE);
			}
		}
	}
	errmsgno(EX_BAD, "Bad timespec '%s' for '%s' in extended header.\n",
		arg, keyword);
	return (FALSE);
}

/*
 * get read access time from extended header
 */
/* ARGSUSED */
LOCAL void
get_atime(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (get_xtime(keyword, arg, &info->f_atime, &info->f_ansec))
		info->f_xflags |= XF_ATIME;
}

/*
 * get inode status change time from extended header
 */
/* ARGSUSED */
LOCAL void
get_ctime(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (get_xtime(keyword, arg, &info->f_ctime, &info->f_cnsec))
		info->f_xflags |= XF_CTIME;
}

/*
 * get modification time from extended header
 */
/* ARGSUSED */
LOCAL void
get_mtime(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (get_xtime(keyword, arg, &info->f_mtime, &info->f_mnsec))
		info->f_xflags |= XF_MTIME;
}

/*
 * generic function to read args that hold decimal numbers
 */
LOCAL BOOL
get_number(keyword, arg, llp)
	char	*keyword;
	char	*arg;
	Llong	*llp;
{
	Llong	ll;
	char	*p;

	p = astollb(arg, &ll, 10);
	if (*p == '\0') {		/* number read correctly */
		*llp = ll;
		return (TRUE);
	}
	errmsgno(EX_BAD, "Bad number '%s' for '%s' in extended header.\n",
		arg, keyword);
	return (FALSE);
}

/*
 * get user id (if > 2097151)
 */
/* ARGSUSED */
LOCAL void
get_uid(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_xflags |= XF_UID;
		info->f_uid = ll;
	}
}

/*
 * get group id (if > 2097151)
 */
/* ARGSUSED */
LOCAL void
get_gid(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_xflags |= XF_GID;
		info->f_gid = ll;
	}
}

/*
 * Space for returning user/group names.
 */
LOCAL	Uchar	_uname[MAX_UNAME+1];
LOCAL	Uchar	_gname[MAX_UNAME+1];

/*
 * get user name (if name length is > 32 chars or if contains non ASCII chars)
 */
/* ARGSUSED */
LOCAL void
get_uname(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (strlen(arg) > MAX_UNAME)
		return;
	if (from_utf8(_uname, (Uchar *)arg)) {
		info->f_xflags |= XF_UNAME;
		info->f_uname = (char *)_uname;
		info->f_umaxlen = strlen((char *)_uname);
	} else {
		bad_utf8(keyword, arg);
	}
}

/*
 * get group name (if name length is > 32 chars or if contains non ASCII chars)
 */
/* ARGSUSED */
LOCAL void
get_gname(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (strlen(arg) > MAX_UNAME)
		return;
	if (from_utf8(_gname, (Uchar *)arg)) {
		info->f_xflags |= XF_GNAME;
		info->f_gname = (char *)_gname;
		info->f_gmaxlen = strlen((char *)_gname);
	} else {
		bad_utf8(keyword, arg);
	}
}

/*
 * get path (if name length is > 100-255 chars or if contains non ASCII chars)
 */
/* ARGSUSED */
LOCAL void
get_path(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (strlen(arg) > PATH_MAX)
		return;
	if (from_utf8((Uchar *)info->f_name, (Uchar *)arg)) {
		info->f_xflags |= XF_PATH;
	} else {
		bad_utf8(keyword, arg);
	}
}

/*
 * get link path (if name length is > 100 chars or if contains non ASCII chars)
 */
/* ARGSUSED */
LOCAL void
get_lpath(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
extern	char	longlinkname[];

	if (strlen(arg) > PATH_MAX)
		return;
	if (from_utf8((Uchar *)longlinkname, (Uchar *)arg)) {
		info->f_xflags |= XF_LINKPATH;
		info->f_lname = longlinkname;
	} else {
		bad_utf8(keyword, arg);
	}
}

/*
 * get size (usually when size is > 8 GB)
 */
/* ARGSUSED */
LOCAL void
get_size(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_xflags |= XF_SIZE;
		info->f_size = (off_t)ll;
	}
}

/*
 * get major device number (always vendor unique)
 */
/* ARGSUSED */
LOCAL void
get_major(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_xflags |= XF_DEVMAJOR;
		info->f_rdevmaj = ll;
	}
}

/*
 * get minor device number (always vendor unique)
 */
/* ARGSUSED */
LOCAL void
get_minor(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_xflags |= XF_DEVMINOR;
		info->f_rdevmin = ll;
	}
}

/*
 * get device number of device containing FS (vendor unique)
 */
/* ARGSUSED */
LOCAL void
get_dev(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_dev = ll;
	}
}

/*
 * get inode number for this file (vendor unique)
 */
/* ARGSUSED */
LOCAL void
get_ino(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_ino = ll;
	}
}

/*
 * get link count for this file (vendor unique)
 */
/* ARGSUSED */
LOCAL void
get_nlink(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	Llong	ll;

	if (get_number(keyword, arg, &ll)) {
		info->f_nlink = ll;
	}
}

/*
 * get tar file type or real file type for this file (vendor unique)
 */
/* ARGSUSED */
LOCAL void
get_filetype(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	int	i;

	for (i=0; i < XT_BAD; i++) {
		if (streql(xttoname_tab[i], arg))
			break;
	}
	if (i >= XT_BAD)
		return;

	if (keyword[7] == 'f') {		/* "SCHILY.filetype" */
		info->f_rxftype = i;
		info->f_filetype = XTTOST(info->f_rxftype);
		info->f_type = XTTOIF(info->f_rxftype);	
	} else {				/* "SCHILY.tarfiletype" */
		info->f_xftype = i;
	}
}

#ifdef	USE_ACL

/*
 * XXX acl_access_text/acl_default_text are a bad idea. (see acl_unix.c)
 */
LOCAL char acl_access_text[PATH_MAX+1];
LOCAL char acl_default_text[PATH_MAX+1];

/* ARGSUSED */
LOCAL void
get_acl_access(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (strlen(arg) > PATH_MAX)	/* XXX We should use dynamic strings */
		return;
	if (from_utf8((Uchar *)acl_access_text, (Uchar *)arg)) {
		info->f_xflags |= XF_ACL_ACCESS;
		info->f_acl_access = acl_access_text;
	} else {
		bad_utf8(keyword, arg);
	}
}

/* ARGSUSED */
LOCAL void
get_acl_default(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	if (strlen(arg) > PATH_MAX)	/* XXX We should use dynamic strings */
		return;
	if (from_utf8((Uchar *)acl_default_text, (Uchar *)arg)) {
		info->f_xflags |= XF_ACL_DEFAULT;
		info->f_acl_default = acl_default_text;
	} else {
		bad_utf8(keyword, arg);
	}
}

#endif  /* USE_ACL */

#ifdef	USE_FFLAGS

/* ARGSUSED */
LOCAL void
get_xfflags(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	texttoflags(info, arg);
	info->f_xflags |= XF_FFLAGS;
}

#endif	/* USE_FFLAGS */

LOCAL void
get_dir(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
	/*
	 * The non UTF-8 string is shorter so we convert in place.
	 */
	from_utf8l((Uchar *)arg, (Uchar *)arg, &len);
	info->f_dir = arg;
	info->f_dirlen = len;
}

/*
 * Dummy 'get' function used for all fields that we don't yet understand or
 * fields that we indent to ignore.
 */
/* ARGSUSED */
LOCAL void
get_dummy(info, keyword, arg, len)
	FINFO	*info;
	char	*keyword;
	char	*arg;
	int	len;
{
}

LOCAL void
bad_utf8(keyword, arg)
	char	*keyword;
	char	*arg;
{
	errmsgno(EX_BAD, "Bad UTF-8 arg '%s' for '%s' in extended header.\n",
		arg, keyword);
}
